/*
* Copyright (c) 2008, Konstantin Tretyakov,
* University of Tartu, Estonia.
* Institute of Computer Science, Bioinformatics Group.
* http://biit.cs.ut.ee, http://phd.kt.pri.ee.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Tartu nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY KONSTANTIN TRETYAKOV ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL KONSTANTIN TRETYAKOV BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
// This is version 0.1 of the code.
// Last modification: 18.09.2008.

// A structure describing a motif
function MotifInfo(name, description, color) {
	this.name = name;
	this.description = description;
	this.color = color;
}

// The sequence just acts as a coordinate system with range fromCoord ... toCoord.
// 0 should correspond to TSS, negative coordinates - upstream.
function SequenceInfo(name, description, fromCoord, toCoord) {
	this.name = name;
	this.description = description;
	this.fromCoord = fromCoord;
	this.toCoord = toCoord;
}

// A site: an instance of a motif on a sequence
function SiteInfo(siteId, motifId, sequenceId, fromCoord, toCoord, score) {
	this.siteId = siteId; // We need a unique site id for convenience
	this.motifId = motifId;
	this.sequenceId = sequenceId;
	this.fromCoord = fromCoord;
	this.toCoord = toCoord;
	this.score = score;
}

// -------------------------------------------- MotifDataProvider ------------------------------------------- //
// An interface providing data about sequences, motifs and sites
// This is just a sample. It should be reimplemented for "proper" use.
function MotifDataProvider() {
	this.motifs = [new MotifInfo("MOTIF1", "Binding site for XX", "red"),
				   new MotifInfo("Motif2", "Binding site for yY", "green"),
				   new MotifInfo("Motif3", "Binding site for Zz", "blue"),
				   new MotifInfo("Motif4", "Binding site for Www", "magenta")];
	this.sequences = [new SequenceInfo("GENE_X", "An important gene responsible for the X functionality", -600, 20),
					  new SequenceInfo("GENE_Y", "An important gene responsible for the Y functionality", -500, 10),
					  new SequenceInfo("GENE_Z", "An important gene responsible for the Z functionality", -700, 30)];
	this.sites = [];
	for(var i = 0; i < 30; i++) {
		var seq   = Math.floor(Math.random()*3);
		var seqlen = this.sequences[seq].toCoord - this.sequences[seq].fromCoord - 30;
		var mtf   = Math.floor(Math.random()*4);
		var sfrom = this.sequences[seq].fromCoord + Math.floor(Math.random()*seqlen);
		var sto   = sfrom + Math.floor(Math.random()*30);
		var score = Math.random();
		this.sites[this.sites.length] = new SiteInfo(i, mtf, seq, sfrom, sto, score);
	}
}

MotifDataProvider.prototype.getSequences = function() {
	return this.sequences;
}

MotifDataProvider.prototype.getSequenceMinCoord = function() {
	return -700;
}

MotifDataProvider.prototype.getSequenceMaxCoord = function() {
	return 30;
}

MotifDataProvider.prototype.getMotifs = function() {
	return this.motifs;
}

// List sites with minScore <= score < maxScore. Allows for AJAX here.
MotifDataProvider.prototype.getSites = function(minScore, maxScore) {
	var r = [];
	for (var i = 0; i < this.sites.length; i++){
		if (this.sites[i].score >= minScore && this.sites[i].score < maxScore) r[r.length] = this.sites[i];
	}
	return r;
}

MotifDataProvider.prototype.getSiteMaxScore = function() {
	return 1;
}

MotifDataProvider.prototype.getSiteMinScore = function() {
	return 0;
}

var SampleMotifDataProvider = MotifDataProvider;

// -------------------------------------------- MotifViewer ------------------------------------------- //
// Constructor requires an id of the div element holding the whole thing and a MotifDataProvider instance.
function MotifViewer(divId, dataProvider) {
	this.divId = divId;
	this.dataProvider = dataProvider;
	this.currentThreshold = null;
	this.motifsDisabled = {};
}

// Initializes the view "from scratch": creates div's responsible for sequences.
// If the sequences/motifs change in the data provider, call this function again.
MotifViewer.prototype.initView = function() {
	this.div = getElement(this.divId);
	this.xMin   = this.dataProvider.getSequenceMinCoord();
	this.xRange = this.dataProvider.getSequenceMaxCoord() - this.xMin;
	this.currentThreshold = null;
	this.motifsDisabled = {};
	var mv = this;
	
	function makeSequenceDiv(seqIndex, seqInfo) {
		var description = "Gene: " + seqInfo.name + "<br/>Description: " + seqInfo.description + "<br/>Range: [" + seqInfo.fromCoord + " .. " + seqInfo.toCoord + "]";
		return DIV(null, 
					DIV({"class": "MV_seqtitle",
					     "onmouseover": "Tip('" + description + "')",
						 "onmouseout": "UnTip()"}, 
						 seqInfo.name),
					DIV({"class": "MV_seqcontainer"},
						DIV({"class": "MV_sequence", 
							"id": mv.divId + "_seq_" + seqIndex,
							"style": "left: "  + 100.0*(seqInfo.fromCoord - mv.xMin)/mv.xRange + "%;" +
									"width: " + 100.0*(seqInfo.toCoord - seqInfo.fromCoord)/mv.xRange + "%;"},
								DIV({"class": "MV_seqmark",
									 "style": "left: "  + 100.0*(-seqInfo.fromCoord)/(seqInfo.toCoord - seqInfo.fromCoord) + "%;"}, null)
							)
						)
					);
	}
	
	// Function actual code starts here
	var seqNodes = [];
	var seqs  = this.dataProvider.getSequences();
	for(var i = 0; i < seqs.length; i++) {
		seqNodes[i] = makeSequenceDiv(i, seqs[i]);
	}
	seqNodes[seqNodes.length] = DIV({"style": "clear: both"},"");
	replaceChildNodes(this.div, seqNodes);
}

// This should be bound to the threshold scrollbar changes
MotifViewer.prototype.updateThreshold = function(newThreshold) {
	var oldThreshold = this.currentThreshold;
	if (oldThreshold == null) {
		oldThreshold = this.dataProvider.getSiteMaxScore();
	}
	
	var mv = this;
	function makeSiteDiv(siteInfo, sequences, motifs) {
		var seq = sequences[siteInfo.sequenceId];
		var mot = motifs[siteInfo.motifId];
		var seqLen = seq.toCoord - seq.fromCoord;
		var siteLen = siteInfo.toCoord - siteInfo.fromCoord;
		var description = "Motif: " + mot.name + "<br/>Description: " + mot.description + "<br/>Site position: [" + siteInfo.fromCoord + " .. " + siteInfo.toCoord + "]<br/>Score: " + siteInfo.score;
		var visibility = "visible;";
		if (mv.motifsDisabled[siteInfo.motifId] == true) visibility = "hidden;";
		return DIV({"id":    mv.divId + "_site_" + siteInfo.siteId,
		            "class": "MV_site MV_motif_" + siteInfo.motifId,
					"onmouseover" : "Tip('" + description + "');",
					"onmouseout"  : "UnTip()",
					"style": "left: "  + 100.0*(siteInfo.fromCoord - seq.fromCoord)/seqLen + "%;" + 
					         "width: " + 100.0*siteLen/seqLen + "%;" +
							 "background-color: " + mot.color + ";" +
							 "visibility: " + visibility }, null);
	}
	
	// Actual code starts here
	if (newThreshold > oldThreshold) {
		// Remove sites
		var sitesToRemove = this.dataProvider.getSites(oldThreshold, newThreshold);
		for (var i = 0; i < sitesToRemove.length; i++) {
			var siteEl = getElement(this.divId + "_site_" + sitesToRemove[i].siteId);
			removeElement(siteEl);
		}
	}
	else if (newThreshold < oldThreshold) {
		// Add new sites to the picture
		var sitesToAdd = this.dataProvider.getSites(newThreshold, oldThreshold);
		for (var i = 0; i < sitesToAdd.length; i++) {
			var seqId = sitesToAdd[i].sequenceId;
			var seqEl = getElement(this.divId + "_seq_" + seqId);
			appendChildNodes(seqEl, makeSiteDiv(sitesToAdd[i], this.dataProvider.getSequences(), this.dataProvider.getMotifs()));
		}
	}
	this.currentThreshold = newThreshold;
}

// Should be bound to the motifSelector's motif toggle method
MotifViewer.prototype.toggleMotif = function(motifId) {
	if (this.motifsDisabled[motifId] == true) {
		this.motifsDisabled[motifId] = false;
	}
	else {
		this.motifsDisabled[motifId] = true;
	}
	// Find all <divs> with class MV_motif_<id> and show them
	var elms = getElementsByClassName(this.div, "DIV", "MV_motif_" + motifId);
	var newVisibility = (this.motifsDisabled[motifId] == true ? "hidden" : "visible");
	for (var i = 0; i < elms.length; i++) elms[i].style.visibility = newVisibility;
}

// -------------------- Motif selector ----------------- //
// A motif selector is a visual control allowing to show/hide motifs
function MotifSelector(divId, motifViewer) {
	this.divId = divId;
	this.motifViewer = motifViewer;
	this.motifs = motifViewer.dataProvider.getMotifs();
}

MotifSelector.prototype.initView = function () {
	this.div = getElement(this.divId);
	function makeMotifDiv(motifId, motifInfo, motifSelector) {
		return DIV({"class": "MV_motif",
					"onclick": function() { motifSelector.toggleMotif(motifId); }}, 
					DIV({"class": "MV_motif_button", 
						 "id" : "MV_motif_button_" + motifId,
					     "style": "background-color: " + motifInfo.color + ";"}, ""), SPAN(null, motifInfo.name)
				); 
	}
	var motifNodes = [];
	for(var i = 0; i < this.motifs.length; i++) {
		motifNodes[i] = makeMotifDiv(i, this.motifs[i], this);
	}
	motifNodes[motifNodes.length] = DIV({"style": "clear: both"},"");
	replaceChildNodes(this.div, motifNodes);
}

MotifSelector.prototype.toggleMotif = function(motifId) {
	this.motifViewer.toggleMotif(motifId);
	if (this.motifViewer.motifsDisabled[motifId] == true) {
		getElement("MV_motif_button_" + motifId).style.backgroundColor = "transparent";
	}
	else {
		getElement("MV_motif_button_" + motifId).style.backgroundColor = this.motifs[motifId].color;
	}
}


// -------------------------- Helper -------------------- //
function getElementsByClassName(oElm, strTagName, strClassName){
	var arrElements = (strTagName == "*" && document.all)? document.all : oElm.getElementsByTagName(strTagName);
	var arrReturnElements = new Array();
	strClassName = strClassName.replace(/\-/g, "\\-");
	var oRegExp = new RegExp("(^|\\s)" + strClassName + "(\\s|$)");
	var oElement;
	for(var i=0; i<arrElements.length; i++){
		oElement = arrElements[i];
		if(oRegExp.test(oElement.className)){
			arrReturnElements.push(oElement);
		}
	}
	return (arrReturnElements)
}
